"""
zmail.server
~~~~~~~~~~~~
This module provides a MailServer object to communicate with mail server.
"""

import smtplib
import poplib
import logging

from .utils import make_iterable
from .message import parse_header, mail_encode, mail_decode
from .info import get_supported_server_info
from .settings import __level__, __local__, __protocol_log__

logger = logging.getLogger('zmail')


class MailServer:
    """This object communicate with server directly."""

    def __init__(self, user, password, smtp_host=None, smtp_port=None, pop_host=None, pop_port=None, smtp_ssl=None,
                 pop_ssl=None):
        self.user = user
        self.password = password

        self.smtp_host = smtp_host
        self.smtp_port = smtp_port
        self.smtp_ssl = smtp_ssl

        self.pop_host = pop_host
        self.pop_port = pop_port
        self.pop_ssl = pop_ssl

    def send_mail(self, recipients, message, timeout=60):
        """"Send email."""

        _message_add_from(message, self.user)

        recipients = make_iterable(recipients)

        host, port, ssl = get_supported_server_info(self.user, 'smtp')
        host = self.smtp_host if self.smtp_host else host
        port = self.smtp_port if self.smtp_port else port
        ssl = self.pop_ssl if self.pop_ssl is not None else ssl

        logger.info('Prepare login into {}:{} ssl:{}.'.format(host, port, ssl))

        server = SMTPServer(host, port, self.user, self.password)

        message = mail_encode(message)

        if ssl:
            server.send_ssl(recipients, message, timeout)
        else:
            server.send(recipients, message, timeout)

    def stat(self):
        """Get mailbox status."""
        server = self._init_pop3()
        status = server.stat()
        server.logout()

        return status

    def get_mail(self, which):
        """Get a mail from mailbox."""
        server = self._init_pop3()

        mail = server.get_mail(which)

        server.logout()

        return mail_decode(mail, which)

    def get_mails(self, subject=None, after=None, before=None, sender=None):
        """Get a list of mails from mailbox."""
        info = self.get_info()
        mail_id = []
        result = []

        for mail in info:
            if self._match(mail, subject, after, before, sender):
                mail_id.append(int(mail['id']))

        server = self._init_pop3()

        for i in mail_id:
            mail = server.get_mail(i)
            result.append(mail_decode(mail, i))
        server.logout()

        return result

    def get_latest(self):
        """Get latest mail in mailbox."""
        server = self._init_pop3()

        latest_num = server.stat()[0]
        mail = server.get_mail(latest_num)

        server.logout()

        return mail_decode(mail, latest_num)

    def get_info(self):
        """Get all mails information.include(subject,from,to,date)"""
        server = self._init_pop3()

        result = server.get_info()

        server.logout()

        return result

    def _init_pop3(self):
        """Initiate POP3 server."""
        host, port, ssl = get_supported_server_info(self.user, 'pop3')
        host = self.pop_host if self.pop_host else host
        port = self.pop_port if self.pop_port else port
        ssl = self.pop_ssl if self.pop_ssl is not None else ssl

        logger.info('Prepare login into {}:{} ssl:{}.'.format(host, port, ssl))
        server = POP3Server(host, port, self.user, self.password, ssl)

        return server

    @staticmethod
    def _match(mail, subject=None, after=None, before=None, sender=None):
        """Match all conditions."""
        _subject = mail['subject']
        _date = mail['date'].split(' ')[0]
        _sender = mail['from']

        _time = tuple(map(lambda x: int(x), _date.split('-')))

        if subject and _subject.find(subject) == -1:
            return False

        if sender and _sender.find(sender) == -1:
            return False

        if before:
            # bf_year, bf_month, bf_day.
            before_time = tuple(map(lambda x: int(x), before.split('-')))
            for i in range(0, 3):
                if before_time[i] < _time[i]:
                    return False
                elif before_time[i] > _time[i]:
                    break
        if after:
            after_time = tuple(map(lambda x: int(x), after.split('-')))
            for i in range(0, 3):
                if after_time[i] > _time[i]:
                    return False
                elif after_time[i] < _time[i]:
                    break
        return True


class SMTPServer:
    def __init__(self, host, port, user, password):
        self.host = host
        self.port = port
        self.user = user
        self.password = password

    def send_ssl(self, recipients, message, timeout):
        with smtplib.SMTP_SSL(self.host, self.port, __local__, timeout=timeout) as server:
            if __protocol_log__:
                server.set_debuglevel(__level__)
            server.login(self.user, self.password)
            for recipient in recipients:
                _message_add_to(message, recipient)
                server.sendmail(self.user, recipient, message.as_string())

    def send(self, recipients, message, timeout, tls=True):
        with smtplib.SMTP(self.host, self.port, __local__, timeout=timeout) as server:
            if __protocol_log__:
                server.set_debuglevel(__level__)
            if tls:
                server.ehlo()
                server.starttls()
                server.ehlo()
            server.login(self.user, self.password)
            for recipient in recipients:
                _message_add_to(message, recipient)
                server.sendmail(self.user, recipient, message.as_string())


class POP3Server:
    def __init__(self, host, port, user, password, ssl=True, tls=True):
        self.user = user
        self.password = password

        self.pop3 = poplib.POP3_SSL(host, port) if ssl else poplib.POP3(host, port)

        if __protocol_log__:
            self.pop3.set_debuglevel(__level__)

        if tls and ssl is False:
            self.pop3.stls()
        self.login()

    def get_info(self):
        """Get all mails info. The result is the form [header_as_dict,...,] """
        num = self.stat()[0]
        result = []
        for count in range(1, num + 1):
            header = self.get_header(count)
            header_as_dict = parse_header(header)
            # Add mail id.
            header_as_dict['id'] = count
            result.append(header_as_dict)
        return result

    def login(self):
        self.pop3.user(self.user)
        self.pop3.pass_(self.password)

    def logout(self):
        self.pop3.quit()

    def stat(self):
        """Get mailbox status. The result is a tuple of 2 integers: (message count, mailbox size)."""
        return self.pop3.stat()

    def get_mail(self, which):
        """Get a mail by its id."""

        mail = self.pop3.retr(which)[1]

        return mail

    def get_header(self, which):
        """Use 'top' to get mail headers."""
        return self.pop3.top(which, 0)[1]


def _message_add_from(message, sender_address):
    """Add 'From' header to avoid server reject the mail."""
    message['From'] = '{}<{}>'.format(sender_address.split('@')[0], sender_address)


def _message_add_to(message, recipient_address):
    """Add 'To' header to avoid server reject the mail."""
    message['To'] = '{}<{}>'.format(recipient_address.split('@')[0], recipient_address)



"""
zmail.info
~~~~~~~~~~~~
This module provide supported server information.


'Server_provider_address':{
    'protocol':('protocol_server_address', port, ssl),
}
"""
supported_server = {
    '163.com': {
        'smtp': ('smtp.163.com', 994, True),
        'pop3': ('pop.163.com', 995, True),
        'imap': ('imap.163.com', 993, True)

    },
    '126.com': {
        'smtp': ('smtp.126.com', 994, True),
        'pop3': ('pop.126.com', 995, True),
        'imap': ('imap.126.com', 993, True)

    },
    'yeah.net': {
        'smtp': ('smtp.yeah.net', 994, True),
        'pop3': ('pop.yeah.net', 995, True),
        'imap': ('imap.yeah.net', 993, True)

    },
    'qq.com': {
        'smtp': ('smtp.qq.com', 465, True),
        'pop3': ('pop.qq.com', 995, True),
    },
    'gmail.com': {
        'smtp': ('smtp.gmail.com', 587, False),
        'pop3': ('pop.gmail.com', 995, True),
    },
    'sina.com': {
        'smtp': ('smtp.sina.com', 465, True),
        'pop3': ('pop.sina.com', 995, True),
    },
    'outlook.com': {
        'smtp': ('smtp-mail.outlook.com', 587, False),
        'pop3': ('pop.outlook.com', 995, True),
    },

}


def get_supported_server_info(mail_address, protocol):
    """Use user address to get server address and port."""
    provider = mail_address.split('@')[1]

    if provider in supported_server:
        server_info = supported_server[provider]
        if protocol in server_info:
            return server_info[protocol]

    if protocol == 'smtp':
        return 'smtp.' + provider, 465, True
    elif protocol == 'pop3':
        return 'pop3.' + provider, 995, True



"""
zmail.message
~~~~~~~~~~~~
This module provides functions to handles MIME object.
"""
import os
import re
import mimetypes
import logging
import quopri
import base64

from email.mime.text import MIMEText
from email.mime.image import MIMEImage
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.encoders import encode_base64
from email.header import decode_header

from .utils import type_check, get_abs_path, make_iterable, bytes_to_string, str_decode

logger = logging.getLogger('zmail')


def mail_encode(message):
    """Convert a dict to a MIME obj."""
    logger.info('Encoding mail to MIME obj.')

    # Type check.
    type_check(dict, message)

    # Create MIMEMultipart.
    msg = MIMEMultipart()

    # Set basic email elements.
    for k, v in message.items():
        if k.capitalize() in ('From', 'To', 'Subject') and v:
            msg[k.capitalize()] = v

    # Set extra parameters.
    for k in message:
        if k.capitalize() not in ('From', 'To', 'Subject') and k not in ('attachments', 'content'):
            msg[k] = message[k]

    # Set mail content.
    if 'content_html' in message:
        msg.attach(MIMEText('%s' % message['content_html'], 'html', 'utf-8'))
    if 'content' in message:
        msg.attach(MIMEText('%s' % message['content'], 'plain', 'utf-8'))

    # Set attachments.
    if 'attachments' in message and message['attachments']:
        attachments = make_iterable(message['attachments'])
        for attachment in attachments:
            logger.info('Loading %s', attachment)
            attachment = get_abs_path(attachment)
            part = _get_attachment_part(attachment)
            msg.attach(part)

    return msg


def mail_decode(mail_as_bytes, which=1):
    """Convert a MIME bytes to dict."""
    result = {}

    # Init.
    for i in ('content', 'contents', 'attachments',):
        result.setdefault(i)
    content = None
    attachments = None

    # Parse headers.
    headers = parse_header(mail_as_bytes, 'boundary')
    result['header'] = headers

    # Add id.
    result['id'] = which

    # Try get ('subject', 'date', 'from', 'to', 'content-type', 'boundary')
    _basic = {'subject': 'Subject', 'date': 'Date', 'from': 'From', 'to': 'To',
              'content-type': 'Content-Type', 'boundary': 'boundary'}
    for k, v in _basic.items():
        try:
            result[k] = headers[v]
        except KeyError:
            result[k] = None
            logger.warning('Can not get element %s in this mail.' % v)
    # Get body.
    body_as_bytes = None
    for line in mail_as_bytes:
        if line in (b'\r\n', b''):
            body_as_bytes = mail_as_bytes[mail_as_bytes.index(line):]
    assert body_as_bytes is not None, 'Can not divide header and body'

    # Decode multi-part mail or one-part mail.
    if result['boundary']:
        logger.info('Decoding multi-part body.id:{},{}'.format(result['id'], result['boundary']))
        body, content, attachments = multiple_part_decode(body_as_bytes, result['boundary'])
    else:
        logger.info('Decoding one-part body.id:{}'.format(result['id']))
        body = one_part_decode(body_as_bytes, result['content-type'])
        if result['content-type'] == 'text/plain':
            result['content'] = body

    result['contents'] = body

    # Add content|attachment if possible

    if content:
        result['content'] = content
    if attachments:
        result['attachments'] = attachments

    result['raw'] = mail_as_bytes

    return result


def parse_header(mail_as_bytes, *args):
    """Parse mail header then return a dictionary include basic elements in email."""
    email_headers = {}

    # Parse headers.
    extra_header = {}
    for i in args:
        extra_header[i] = None
    extra_pattern = rb"""[\s]?=["'\s]?([^"';\r\n\s]+)["';\r\n\s]?"""

    part = b''
    header = b''

    for line in mail_as_bytes:
        # Get extra_header if possible.
        for k, v in extra_header.items():
            if v is None and re.search(bytes(k, encoding='utf8') + bytes(extra_pattern), line):
                extra_header[k] = re.search(bytes(k, encoding='utf8') + bytes(extra_pattern), line).group(1)

        # Parsing header.
        if re.fullmatch(rb'([0-9a-zA-Z-]+):\s?([^\r\n]+)(\r\n)?', line):
            # Line is a header.
            if part and header:
                email_headers[header.decode()] = part
            header, part = re.search(rb'([0-9a-zA-Z-]+):\s?([^\r\n]+)(\r\n)?', line).group(1, 2)
        elif re.fullmatch(rb'[\t]?([^\r\n]+)(\r\n)?', line):
            # Line is a part of header.
            _part = re.search(rb'[\t]?([^\r\n]+)(\r\n)?', line).group(1)
            part += _part
        elif line in (b'\r\n', b''):
            # Mail header boundary.
            if part and header:
                email_headers[header.decode()] = part
            break
        else:
            raise Exception('Parsing error:%s' % line)

    # Decode headers.
    for k, v in email_headers.items():
        groups = decode_header(v.decode())
        string = ''
        for group in groups:
            _string, charset = group
            string += str_decode(_string, charset)
        email_headers[k] = string

    # Add extra header if possible.
    email_headers = {**email_headers, **extra_header}

    return email_headers


def _get_attachment_part(file):
    """According to file-type return a prepared attachment part."""
    name = os.path.split(file)[1]
    file_type = mimetypes.guess_type(name)[0]

    if file_type is None:
        logger.warning('Could not guess %s type, use application type instead.', file)
        file_type = 'application/octet-stream'

    main_type, sub_type = file_type.split('/')

    if main_type == 'text':
        with open(file, 'r') as f:
            part = MIMEText(f.read())
            part['Content-Disposition'] = 'attachment;filename="%s"' % name

    elif main_type in ('image', 'audio'):
        with open(file, 'rb') as f:
            part = MIMEImage(f.read(), _subtype=sub_type) if main_type == 'image' else \
                MIMEAudio(f.read(), _subtype=sub_type)
            part['Content-Disposition'] = 'attachment;filename="%s"' % name
    else:
        with open(file, 'rb') as f:
            part = MIMEBase(main_type, sub_type)
            part.set_payload(f.read())
            part['Content-Disposition'] = 'attachment;filename="%s"' % name
            encode_base64(part)

    return part


def _fmt_date(date_as_bytes):
    """Format mail header Date for humans."""
    date_as_string = date_as_bytes.decode()

    _month = {
        'Jan': 1,
        'Feb': 2,
        'Mar': 3,
        'Apr': 4,
        'May': 5,
        'Jun': 6,
        'Jul': 7,
        'Aug': 8,
        'Sep': 9,
        'Oct': 10,
        'Nov': 11,
        'Dec': 12,
    }
    date_list = date_as_string.split(',')
    # week = date_list[0].replace(' ', '')
    date_list = date_list[1].split(' ')

    date_list = list(filter(lambda x: x != '', date_list))
    day = date_list[0]
    month = _month[date_list[1]]
    year = date_list[2]
    times = date_list[3]
    time_zone = date_list[4]

    return '{}-{}-{} {} {}'.format(year, month, day, times, time_zone)


def multiple_part_decode(body_as_bytes, boundary):
    """Convert MIME body to content."""
    # Get part|content|attachments if possible.
    attachments = []
    content = []
    parts = []

    # Divide into multiple parts.
    is_part = False
    part = []
    for i in body_as_bytes:

        if i.find(boundary) > -1:
            is_part = True
            if part:
                parts.append(part)
            part = []
            continue

        if is_part:
            part.append(i)

    # Parse each parts.
    for p in parts:
        # Parse each part's header.
        headers = parse_header(p, 'charset')

        content_type = headers['Content-Type'] if headers.get('Content-Type') else None
        encoding = headers['Content-Transfer-Encoding'] if headers.get('Content-Transfer-Encoding') else None
        disposition = headers['Content-Disposition'] if headers.get('Content-Disposition') else None
        charset = headers['charset'] if headers.get('charset') else None

        # Is attachment.
        if disposition and disposition.find('attachment') == 0:
            attachment = []
            filename = re.search(r"""attachment;[\s]?filename=[\s"']?([^"';]+)""", disposition).group(1)
            attachment.append(filename + ';%s' % content_type)

            # Add body.
            is_body = False
            for line in p:
                if line == b'':
                    is_body = True
                    continue
                if is_body:
                    if encoding.find('base64') > -1:
                        attachment.append(base64.b64decode(line))
                    else:
                        attachment.append(line)
            attachments.append(attachment)

        # Is text.
        elif headers.get('Content-Type'):
            coding = headers['Content-Transfer-Encoding'] if headers.get('Content-Transfer-Encoding') else None
            logger.info('Parse {},coding:{}'.format(headers['Content-Type'], coding))
            is_body = False
            for line in p:
                if line == '':
                    is_body = True
                    continue
                if is_body:
                    if coding.find('base64') > -1:
                        content.append(base64.b64decode(line).decode())
                    else:
                        content.append(line.decode())
        else:
            raise Exception('Multiple part error!')

    return parts, content, attachments


def one_part_decode(body_as_bytes, content_type):
    # [could be improved]
    body_as_string = bytes_to_string(body_as_bytes)
    body = ''

    # Could't decode.
    if content_type != 'text/plain':
        logger.info("Could't decode body, use raw instead.Type:{}".format(content_type))
        body = []
        for i in body_as_string:
            if len(i) == 74:
                body.append(i[:-1])
            else:
                body.append(i)
        return body

    # Type is 'text/plain'.
    for i in body_as_string:
        if len(i) == 74 and i[-1] == '=':
            body += quopri.decodestring(i[:-1]).decode()
        else:
            body += quopri.decodestring(i).decode()
    return body



"""
zmail.api
~~~~~~~~~~~~
This module implements the zmail API.
"""

from .server import MailServer
from .utils import get_attachment, get_html, show, read_eml, save_eml
from .message import mail_decode


def server(user, password, smtp_host=None, smtp_port=None, pop_host=None, pop_port=None, smtp_ssl=None, pop_ssl=None):
    """A shortcut to use MailServer.

    SMTP:
        server.send_mail([recipient,], mail)

    POP3:
        server.get_mail(which)
        server.get_mails(subject, sender, after, before)
        server.get_latest()
        server.get_info()
        server.stat()

    Parse mail:
        server.show(mail)
        server.get_attachment(mail)
    """
    return MailServer(user, password, smtp_host=smtp_host, smtp_port=smtp_port, pop_host=pop_host, pop_port=pop_port,
                      smtp_ssl=smtp_ssl,
                      pop_ssl=pop_ssl)


def decode(mail_as_bytes, which=1):
    return mail_decode(mail_as_bytes, which)



"""
zmail.utils
~~~~~~~~~~~~
This module contains some useful function power zmail.
"""

import os
import sys


def bytes_to_string(bytes_list):
    """Decode a list of bytes-objects to a list of string-objects."""
    return tuple(map(lambda x: x.decode(), bytes_list))


def type_check(_types, *args):
    """Each args must be a type in _types or raise a TypeError."""
    for obj in args:
        if not isinstance(obj, _types):
            raise TypeError('Parameter {} must be {}'.format(obj, _types))


def get_abs_path(file):
    """if the file exists, return its abspath or raise a exception."""
    work_path = os.path.abspath(os.getcwd())
    if os.path.isfile(os.path.join(work_path, file)):
        return os.path.join(work_path, file)
    elif os.path.isfile(file):
        return file
    else:
        raise Exception("The file %s doesn't exist." % file)


def make_iterable(obj):
    """Get a iterable obj."""
    if isinstance(obj, (tuple, list)):
        return obj
    return obj,


def get_attachment(mail, *args):
    """Parsing attachment and save it."""
    names = list(args)
    names.reverse()
    if mail['attachments']:
        for attachment in mail['attachments']:
            info = attachment[0].split(';')
            name = info[0]
            body_type = info[1]
            is_text_file = True if body_type.find('text/plain') > -1 else False

            try:
                name = names.pop()
            except IndexError:
                pass

            # Write file.
            if not is_text_file:
                # Binary file.
                body = b''.join(attachment[1:])
                with open(name, 'wb') as f:
                    f.write(body)
            else:
                # Text file.
                body = tuple(map(lambda x: x.decode() + '\r\n', attachment[1:]))
                with open(name, 'w') as f:
                    f.writelines(body)


def show(mails):
    """Show mails."""
    mails = make_iterable(mails)
    for mail in mails:
        print('-------------------------')
        for k, v in mail.items():
            print(k, v)


def str_decode(text, coding=None):
    """Decode bytes to string."""
    if isinstance(text, str):
        return text
    elif isinstance(text, bytes):
        if coding:
            return text.decode(coding)
        else:
            return text.decode()
    else:
        raise Exception('String decoding error:%s' % text)


def get_html(html_path):
    """Get html content by its path."""
    path = get_abs_path(html_path)

    with open(path, 'r') as f:
        content = f.read()

    return content


def read_eml(path):
    abs_path = get_abs_path(path)
    with open(abs_path, 'rb') as file:
        eml = file.readlines()
    return eml


def save_eml(mail, name=None, path=None):
    file_name = name if name else str(mail['subject'] + '.eml')
    file_path = path if path else os.path.abspath(os.path.dirname(sys.argv[0]))

    file_locate = os.path.join(file_path, file_name)

    with open(file_locate, 'wb+') as f:
        f.writelines(mail['raw'])

    return True



import os
import sys

from setuptools import setup, find_packages

PROJECT_NAME = 'zmail'
MODULE_NAME = 'zmail'

# 'setup.py publish' shortcut.
if sys.argv[-1] == 'publish':
    os.system('python3 setup.py sdist bdist_wheel')
    os.system('twine upload dist/*')
    sys.exit()

setup(
    name='zmail',
    version='0.0.8',

    author='ZYunH',
    author_email='workvl@163.com',

    description='Zmail allows you to send and get emails as possible as it can be in python',
    long_description='Zmail allows you to send and get emails as possible as it can be in python.There is no need to check server address or make your own MIME string.With zmail, you only need to care about your mail content.',
    keywords='email python3 package',

    url='https://github.com/ZYunH/zmail',
    license="MIT Licence",

    platforms='all',

    packages=find_packages(),
    include_package_data=True,

)
